import os

from jamp.jam_builtins import (Builtins, output)
from jamp.expand import expand, expand_lol, flatten, iter_var, lol_get
from jamp.jam_syntax import Arg, Node
from jamp.classes import Rule, State, Exec, Target, UpdatingAction
from typing import Optional, Union
from jamp.pattern import match

builtins = Builtins()
complained_rules = set()

FLOW_BREAK = 1
FLOW_CONTINUE = 2
FLOW_DEBUG = 3


class Result:
    def __init__(self, val):
        self.val = val


def run(state: State, cmds: Union[list, Exec]) -> Optional[int]:
    """Starting point of tasks execution"""

    res = None
    if cmds:
        if isinstance(cmds, Exec):
            return cmds.execute(state)

        for i, ex in enumerate(cmds):
            ret = ex.execute(state)
            if ret:
                # probably we need to break the execution
                if ret in (FLOW_BREAK, FLOW_CONTINUE):
                    return ret
                elif ret == FLOW_DEBUG:
                    import pdb

                    pdb.set_trace()

                if isinstance(ret, Result):
                    res = ret

    return res


def bind_targets(state: State, search_headers="base"):
    """Bind target to actual locations"""

    target: Target = None

    for target in state.targets.values():
        target.bind_location(state)

    db = None
    if search_headers == "ripgrep":
        from jamp.headers import scan_ripgrep_output

        pattern = state.vars.get("HDRPATTERN")
        if pattern:
            db = scan_ripgrep_output(state, pattern[0])
    elif search_headers == "grep":
        from jamp.headers import scan_grep_output

        pattern = state.vars.get("HDRPATTERN")
        if pattern:
            db = scan_grep_output(state, pattern[0])

    if search_headers != "none":
        for target in tuple(state.targets.values()):
            # tuple is because targets dict will change while searching
            if target.boundname:
                target.find_headers(state, db=db)

        # now bind found headers
        for target in state.targets.values():
            target.bind_location(state, strict=True)


class ExecutionError(Exception):
    pass


def check_empty_val(assign_list):
    if assign_list and isinstance(assign_list[0], Arg) and assign_list[0].value == "":
        return True

    return False


def exec_assign(
    state: State, name_arg: Arg, assign_type: str, assign_list=None
) -> None:
    """Assign to a global or local variable"""
    names = expand(state, name_arg)

    for name in names:
        assign = False
        if assign_type == "=":
            # normal assign
            assign = True
        elif assign_type == "?=" or assign_type[0] == "d":
            # if variable is defined here, just skip
            scope = state.vars.get_scope(name)
            if scope is None:
                assign = True
            else:
                val = scope.get(name)
                if not val:
                    assign = True

        elif assign_type == "+=":
            # add to variable
            value = expand(state, assign_list)
            curval = state.vars.get(name)
            if curval:
                state.vars.set(name, list(iter_var(curval)) + list(iter_var(value)))
            elif value:
                # original jam skips setting var if value is empty
                assign = True

        if assign:
            value = expand(state, assign_list, skip_empty=False)
            state.vars.set(name, value)


def exec_assign_on_target(
    state: State, name_arg: Arg, targets, assign_type: str, assign_list
):
    """Bind a variable to specific targets"""
    targets_var = expand(state, targets)

    for target_name in iter_var(targets_var):
        target = Target.bind(state, target_name)

        value = expand(state, assign_list, skip_empty=False)
        names = expand(state, name_arg)
        target_vars = target.vars

        for varname in names:
            if assign_type == "=":
                target_vars[varname] = value
            elif assign_type == "?=":
                if varname not in target_vars:
                    target_vars[varname] = value
            elif assign_type == "+=" and value:
                # note: skip adding if the value is empty
                if varname not in target_vars:
                    target_vars[varname] = value
                else:
                    curval = flatten(target_vars[varname])
                    if curval:
                        target_vars[varname] = list(iter_var(curval)) + list(
                            iter_var(value)
                        )
                    else:
                        target_vars[varname] = value


def exec_break(state: State, arg) -> int:
    return FLOW_BREAK


def exec_continue(state: State, arg) -> int:
    return FLOW_CONTINUE


def exec_return(state: State, retval) -> Result:
    return Result(expand(state, retval))


def exec_local_assign(state: State, names: Union[Arg, list], assign_list):
    names = expand(state, names)
    value = expand(state, assign_list, skip_empty=False)

    for name in names:
        state.vars.set_local(name, value)


def exec_rule_action(state: State, rule: Rule, action_name: str, params: list):
    target_names = lol_get(params, 0)
    source_names = lol_get(params, 1)
    action = state.actions[action_name]

    sources = []
    for source_name in source_names:
        source_t = Target.bind(state, source_name)
        sources.append(source_t)

    bindvars = set()
    if action.bindlist and action.bindlist[1]:
        for arg in action.bindlist[1]:
            bindvars.add(arg.value)

    generated = False
    restat = False
    prev_upd_action = None
    build_targets = []
    linking_targets = []
    for target_name in target_names:
        target = Target.bind(state, target_name)
        if target.generated:
            generated = True

        if target.restat:
            restat = True

        if target.build_step is not None:
            # there is a build step for this target, this target goes there
            prev_upd_action = target.build_step[1]
            linking_targets.append(target)
        else:
            build_targets.append(target)

        for var in bindvars:
            bind_names = state.vars.get(var, on_target=target)
            if bind_names:
                for val in bind_names:
                    # just in case create targets if ACTIONS BIND params are set
                    if val:
                        Target.bind(state, val[0])

    def try_collect(ua, src):
        if ua.action.collect_together and ua.action == action:
            ua.sources += src
            return True

        return False

    if prev_upd_action:
        collected = try_collect(prev_upd_action, sources)
        if not collected:
            for ua in prev_upd_action.next:
                collected = try_collect(ua, sources)
                if collected:
                    break

        if not collected:
            upd_action = UpdatingAction(action, sources)
            upd_action.targets = linking_targets
            upd_action.bindvars = bindvars
            prev_upd_action.link(upd_action)

    if build_targets:
        # one build step, can output several targets
        upd_action = UpdatingAction(action, sources)
        upd_action.targets = build_targets
        upd_action.generator = generated
        upd_action.bindvars = bindvars
        upd_action.restat = restat

        step = (build_targets, upd_action)
        for target in build_targets:
            target.build_step = step
            target.is_output = True

        state.build_steps.append(step)


def exec_one_rule(state: State, name: str, params: list):
    builtin = getattr(builtins, name.lower(), None)
    if builtin:
        return builtin(state, params)

    rule: Rule = state.rules.get(name)
    if rule is None:
        if state.current_rule is not None:
            # look like an actions call inside some rule
            if name in state.actions and name != state.current_rule.name:
                # create an updating action for targets
                exec_rule_action(state, state.current_rule, name, params)
                return

        if name != "Clean":
            # we just ignore clean rules, ninja will do cleaning part
            if name not in complained_rules:
                output(f"jamp: unknown rule {name}")
                complained_rules.add(name)

        return

    # create an updating action for actions block with the same name as the rule
    if name in state.actions:
        exec_rule_action(state, rule, name, params)

    # execute rule commands
    old_params = state.params
    old_rule = state.current_rule

    state.current_rule = rule
    state.params = params
    ret = exec_block(state, rule.commands)
    state.params = old_params
    state.current_rule = old_rule

    return ret


def exec_rule(state: State, name: Arg, args):
    names = expand(state, name)
    params = expand_lol(state, args)

    res = []
    for name in names:
        rule_res: Optional[Result] = exec_one_rule(state, name, params)
        if isinstance(rule_res, Result):
            res += rule_res.val
        elif rule_res == FLOW_DEBUG:
            return rule_res

    if res:
        return Result(res)


def exec_include(state: State, location):
    filenames = expand(state, location)

    for filename in filenames:
        t = Target.bind(state, filename)

        with t.overlay(state):
            t.boundname = t.search(state)

        if state.debug_include:
            print(
                f"Including {t.boundname}, target: {filename}, expanded from: {location}"
            )

        if not t.boundname or not os.path.exists(t.boundname):
            print(f"Include failed on file: {t.boundname}")
            exit(1)
        else:
            jamfiles_target = state.targets.get("jamfiles")
            if jamfiles_target:
                jamfiles_target.depends.add(t)

            with open(t.boundname) as f:
                rules = f.read()
                cmds = state.parse_and_compile(rules, filename=t.boundname)

                state.vars.push()
                run(state, cmds)
                state.vars.pop()


def exec_on_target(state: State, targets, cmds):
    """
    on target statement
    run cmds under target vars influence
    """
    for target_name in expand(state, targets):
        target = state.get_target(target_name)

        with target.overlay(state):
            run(state, cmds)


def exec_rule_on_target(state: State, targets, name, args):
    """
    [ on target rule args ]
    exec rules under target vars influence
    """

    res = []
    for target_name in expand(state, targets):
        target = Target.bind(state, target_name)

        rule_res = None
        with target.overlay(state):
            rule_res = exec_rule(state, name, args)

        if isinstance(rule_res, Result):
            res += rule_res.val

    if res:
        return Result(res)


def exec_return_on_target(state: State, targets, val):
    """
    [ on target return val ]
    return vars target vars influence
    """

    res = []
    for target_name in expand(state, targets):
        target = state.get_target(target_name)
        if target:
            with target.overlay(state):
                res += expand(state, val)

    return Result(res)


def var_bool(var):
    if var is None:
        return False

    if isinstance(var, bool):
        return var
    elif isinstance(var, str):
        return len(var) > 0
    elif isinstance(var, list) and len(var) > 0:
        return len(var[0]) > 0

    return False


def evaluate_expr(state: State, args: tuple):
    match args:
        case (Node.EXPR, arg):
            res = expand(state, arg)
            return res

        case (Node.EXPR_BOP, (left, op, right)):
            if op == "in":
                left = expand(state, left)
                # if len(left):
                right = expand(state, right)
                return set(left).issubset(right)

            left = evaluate_expr(state, left)

            # quick path for AND and OR
            match op:
                case "&&":
                    if not var_bool(left):
                        return False
                case "||":
                    if var_bool(left):
                        return True

            right = evaluate_expr(state, right)
            match op:
                case "=":
                    return left == right
                case ">":
                    return left > right
                case "<":
                    return left < right
                case "!=":
                    return left != right
                case ">=":
                    return left >= right
                case "<=":
                    return left <= right
                case "&&":
                    return var_bool(left) and var_bool(right)
                case "||":
                    return var_bool(left) or var_bool(right)
                case _:
                    raise ExecutionError(f"unexpected binary operation: {op}")
        case (Node.EXPR_UNARY, op, arg):
            arg = evaluate_expr(state, arg)
            match op:
                case "!":
                    return not var_bool(arg)
                case _:
                    raise ExecutionError(f"unexpected unary operation: {op}")
        case (Node.EXPR_BLOCK, block):
            # braced
            return evaluate_expr(state, block)
        case _:
            raise ExecutionError(f"could not match an expression: {args}")


def exec_if(state: State, cond, block, else_block):
    res = var_bool(evaluate_expr(state, cond))

    if res is True:
        return exec_block(state, block)
    else:
        return exec_block(state, else_block)


def exec_while(state: State, cond, block):
    while var_bool(evaluate_expr(state, cond)):
        ret = exec_block(state, block)
        if ret == FLOW_BREAK:
            break
        elif ret == FLOW_CONTINUE:
            continue


def exec_for(state: State, var, items, block):
    items = expand(state, items)
    if isinstance(items, str):
        items = (items,)

    varname = expand(state, var)
    if len(varname) == 0:
        raise ExecutionError(f"got empty argument in for after expanding {var}")

    for item in items:
        state.vars.set(varname[0], [item])
        ret = exec_block(state, block)
        if ret == FLOW_BREAK:
            break
        elif ret == FLOW_CONTINUE:
            continue


def exec_switch(state: State, arg, cases):
    arg = expand(state, arg)

    for pattern, block in cases:
        matched = False
        if not arg and pattern == "*":
            matched = True

        if arg and match(pattern, arg[0]) == 0:
            matched = True

        if matched:
            if len(block) and isinstance(block[0], list):
                return exec_block(state, block[0])
            else:
                return run(state, block)


def exec_block(state, cmds):
    # create a local scope
    state.vars.push()

    ret = run(state, cmds)

    # return the scope
    state.vars.pop()

    return ret



#!/usr/bin/env python3

import sys

from jamp.jam_lexer import Lexer, SCAN_NORMAL, keywords
from jamp.yacc import yacc
from enum import Enum

use_colors = False

if not hasattr(sys, "_called_from_test"):
    try:
        import colorama
        from colorama import Fore, Style

        colorama.init()
        use_colors = True
    except ImportError:
        pass


def highlight(text, arg=False):
    if use_colors and not arg:
        return f"{Fore.YELLOW}{text}{Style.RESET_ALL}"
    if use_colors and arg:
        return f"{Fore.CYAN}{text}{Style.RESET_ALL}"
    else:
        return f"{text}"


class Node(Enum):
    ASSIGN = 2
    CALL = 3
    LOL = 4
    ACTIONS = 5
    CASE = 7
    ARG_ON_TARGET = 8
    EXPR = 9
    FUNC = 10
    FUNC_ON = 11
    SWITCH = 12
    RULE = 13
    WHILE = 14
    IF = 15
    FOR = 16
    BREAK = 17
    RETURN = 18
    CONTINUE = 19
    INCLUDE = 20
    ON_TARGET = 21
    LOCAL = 22
    EFLAG = 23
    BINDLIST = 24
    EXPR_BOP = 26
    EXPR_UNARY = 27
    EXPR_BLOCK = 28
    RETURN_ON = 29

    def __repr__(self):
        return highlight(self.name)


tokens = list(keywords.values()) + ["ARG", "STRING"]


precedence = (
    ("left", "BARBAR", "BAR"),
    ("left", "AMPERAMPER", "AMPER"),
    ("left", "EQUALS", "BANG_EQUALS", "IN"),
    ("left", "LANGLE", "LANGLE_EQUALS", "RANGLE", "RANGLE_EQUALS"),
    ("left", "BANG"),
)


# parser begin
# seq - zero or more rules
# rules - one or more rules
# rule - any one of jam's rules
#   right-recursive so rules execute in order.
def p_seq(p):
    """
    block :
    | rules
    """
    if len(p) > 1:
        p[0] = p[1]
    else:
        p[0] = None


def p_rules(p):
    """
    rules : rule
        | rule rules
    """
    if len(p) == 2:
        p[0] = [p[1]]
    elif len(p) == 3:
        p[2].insert(0, p[1])
        p[0] = p[2]


def p_rules_local(p):
    """
    local : LOCAL list SEMICOLON
        | LOCAL list EQUALS list SEMICOLON
    """
    if len(p) == 4:
        p[0] = (Node.LOCAL, p[2], None)
    else:
        p[0] = (Node.LOCAL, p[2], p[4])


def p_block(p):
    """
    rule : rule_block
    | include
    | rule_call
    | arg_assign
    | arg_on_target
    | break
    | continue
    | return
    | for
    | switch
    | if
    | while
    | ruledef
    | on_target
    | actions
    | local
    """
    p[0] = p[1]


def p_rule_on_arg(p):
    """
    on_target : ON arg rule
    """
    p[0] = (Node.ON_TARGET, p[2], p[3])


def p_rule_call(p):
    """
    rule_call : arg lol SEMICOLON
    """
    p[0] = (Node.CALL, p[1], p[2])


def p_rule_include(p):
    """
    include : INCLUDE list SEMICOLON
    """
    p[0] = (Node.INCLUDE, p[2])


def p_rule_break(p):
    """
    break : BREAK list SEMICOLON
    """
    p[0] = (Node.BREAK, p[2])


def p_rule_continue(p):
    """
    continue : CONTINUE list SEMICOLON
    """
    p[0] = (Node.CONTINUE, p[2])


def p_rule_return(p):
    """
    return : RETURN list SEMICOLON
    """
    p[0] = (Node.RETURN, p[2])


def p_rule_for(p):
    """
    for : FOR ARG IN list LBRACE block RBRACE
    """
    p[0] = (Node.FOR, p[2], p[4], p[6])


def p_rule_switch(p):
    """
    switch : SWITCH list LBRACE cases RBRACE
    """
    p[0] = (Node.SWITCH, p[2], p[4])


def p_rule_while(p):
    """
    while : WHILE expr LBRACE block RBRACE
    """

    p[0] = (Node.WHILE, p[2], p[4])


def p_ruledef(p):
    """
    ruledef : RULE ARG params LBRACE block RBRACE
    """
    p[0] = (Node.RULE, p[2], p[3], p[5])


def p_rule_if(p):
    """
    if : IF expr LBRACE block RBRACE
    | IF expr LBRACE block RBRACE ELSE rule
    """
    if len(p) == 6:
        p[0] = (Node.IF, p[2], p[4])
    else:
        p[0] = (Node.IF, p[2], p[4], p[7])


def p_rule_arg_assign(p):
    """
    arg_assign : arg assign_type list SEMICOLON
    """
    p[0] = (Node.ASSIGN, p[1], p[2], p[3])


def p_rule_arg_on_target(p):
    """
    arg_on_target : arg ON list assign_type list SEMICOLON
    """
    p[0] = (Node.ARG_ON_TARGET, p[1], p[3], p[4], p[5])


def p_rule_actions(p):
    """
    actions : ACTIONS eflags ARG bindlist LBRACE STRING RBRACE
    """
    p[0] = (Node.ACTIONS, p[2], p[3], p[4], p[6])


def p_rule_block(p):
    """
    rule_block : LBRACE block RBRACE
    """
    p[0] = p[2]


def p_assign_type(p):
    """
    assign_type : EQUALS
    | PLUS_EQUALS
    | QUESTION_EQUALS
    | DEFAULT EQUALS
    """
    p[0] = p[1]


def p_expr(p):
    """
    expr : arg
    | expr EQUALS expr
    | expr BANG_EQUALS expr
    | expr LANGLE expr
    | expr LANGLE_EQUALS expr
    | expr RANGLE expr
    | expr RANGLE_EQUALS expr
    | expr AMPER expr
    | expr AMPERAMPER expr
    | expr BAR expr
    | expr BARBAR expr
    | arg IN list
    | BANG expr
    | LPAREN expr RPAREN
    """
    if len(p) == 2:
        p[0] = (Node.EXPR, p[1])
    elif len(p) == 3:
        # unary
        p[0] = (Node.EXPR_UNARY, p[1], p[2])
    elif len(p) == 4:
        if p[1] == "(":
            p[0] = (Node.EXPR_BLOCK, p[2])
        else:
            p[0] = (Node.EXPR_BOP, (p[1], p[2], p[3]))


# cases - action elements inside a 'switch'
# case - a single action element inside a 'switch'
# right-recursive rule so cases can be examined in order.
def p_cases(p):
    """
    cases :
    | case cases
    """
    if len(p) == 1:
        p[0] = []
    else:
        p[2].insert(0, p[1])
        p[0] = p[2]


def p_case(p):
    """
    case : CASE ARG COLON block
    """
    p[0] = (Node.CASE, p[2], p[4])


# params - optional parameter names to rule definition
# right-recursive rule so that params can be added in order.
def p_params(p):
    """
    params :
    | ARG COLON params
    | ARG
    """
    if len(p) > 1:
        p[0] = (Node.PARAMS, p[1], p[3] if len(p) > 2 else None)


# lol - list of lists
#    right-recursive rule so that lists can be added in order.
def p_lol(p):
    """
    lol : list
        | list COLON lol
    """
    if len(p) == 2:
        p[0] = (Node.LOL, p[1])
    else:
        p[0] = (Node.LOL, p[1], *p[3][1:])


# list - zero or more args in a LIST
def p_list(p):
    """
    list : listp
    """
    p[0] = p[1]
    # p.lexer.set_scanmode(SCAN_NORMAL)


# listp - list (in puncutation only mode)
def p_listp(p):
    """
    listp :
        | listp arg
    """
    if len(p) == 1:
        pass
        # p.lexer.set_scanmode(SCAN_PUNCT)
    else:
        if p[1] is None:
            p[0] = [p[2]]
        else:
            p[1].append(p[2])
            p[0] = p[1]


class Arg(object):
    def __init__(self, val):
        self.value = val

    def __repr__(self):
        if self.value:
            return highlight(self.value, arg=True)
        else:
            return highlight('""', arg=True)


# arg - one ARG or function call
def p_arg(p):
    """
    arg : ARG
        | LBRACKET func RBRACKET
    """
    if len(p) == 2:
        p[0] = Arg(p[1])
    else:
        p.lexer.set_scanmode(SCAN_NORMAL)
        p[0] = Arg(p[2])


# func - a function call (inside [])
# This needs to be split cleanly out of 'rule'
def p_func(p):
    """
    func : arg lol
        | ON arg arg lol
        | ON arg RETURN list

    """
    if len(p) == 3:
        p[0] = (Node.FUNC, p[1], p[2])
    elif p[3] == 'return':
        p[0] = (Node.RETURN_ON, p[2], p[4])
    else:
        p[0] = (Node.FUNC_ON, p[2], p[3], p[4])


# eflags - zero or more modifiers to 'executes'
def p_eflags(p):
    """
    eflags :
    | eflags eflag
    """
    if len(p) == 1:
        p[0] = []
    else:
        p[1].append(p[2])
        p[0] = p[1]


# eflag - a single modifier to 'executes'
def p_eflag(p):
    """
    eflag : UPDATED
    | TOGETHER
    | IGNORE
    | QUIETLY
    | PIECEMEAL
    | EXISTING
    | MAXLINE ARG
    """
    if len(p) == 2:
        p[0] = (Node.EFLAG, p[1], None)
    else:
        p[0] = (Node.EFLAG, p[1], p[2])


# bindlist - list of variable to bind for an action
def p_bindlist(p):
    """
    bindlist :
    | BIND list
    """
    if len(p) > 1:
        p[0] = (Node.BINDLIST, p[2])
    else:
        p[0] = (Node.BINDLIST, None)


def p_error(p):
    parts = []
    if not p:
        print("Error, but no info, check Jamfile for correct EOF")
        return

    if p.lexer and p.lexer.filename:
        parts.append(p.lexer.filename)

    if p.lexer:
        parts.append(str(p.lexer.current_lineno()))

    if p and p.value is not None:
        print(f"Syntax error at {p.value!r} at line {':'.join(parts)}")
    else:
        print("Syntax error")


def parse(text: str, filename: str = None):
    parser = yacc()
    lexer = Lexer(filename=filename)
    lexer.input(text)
    return parser.parse(lexer=lexer)


def parse_file(fn: str):
    parser = yacc()

    with open(fn) as f:
        data = f.read()

    lexer = Lexer()
    lexer.input(data)
    return parser.parse(lexer=lexer)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        prog="jam parser", description="Checks jam grammar"
    )
    parser.add_argument("--tokens", action="store_true", help="print tokens")
    parser.add_argument("--ast", action="store_true", help="print AST")
    parser.add_argument("filename")
    args = parser.parse_args()

    parser = yacc(debug=True)
    lexer = Lexer()
    with open(args.filename) as f:
        data = f.read()

    lexer.input(data)

    if args.tokens:
        while True:
            tok = lexer.token()
            if not tok:
                break  # No more input

            print(tok)

    if args.ast:
        lexer.restart()
        ast = parser.parse(data, lexer=lexer, tracking=True)
        print(ast)



from jamp.jam_syntax import Node, Arg
from jamp.classes import Rule, Exec, Actions, State
from jamp import executors


class CompilerError(Exception):
    pass


def compile(state: State, node: tuple | list):
    match node:
        case Arg(value=(Node.FUNC, *r)):
            return Exec(executors.exec_rule, r)
        case Arg(value=(Node.FUNC_ON, *r)):
            return Exec(executors.exec_rule_on_target, r)
        case Arg(value=(Node.RETURN_ON, targets, val)):
            return Exec(
                executors.exec_return_on_target,
                (
                    compile(state, targets),
                    compile(state, val),
                ),
            )
        case (Node.RULE, *r):
            return Exec(compile_rule, r)
        case (Node.ASSIGN, *r):
            return Exec(executors.exec_assign, compile(state, r))
        case (Node.ARG_ON_TARGET, target, var, assign_type, assign_list):
            return Exec(
                executors.exec_assign_on_target,
                (
                    (compile(state, target)),
                    compile(state, var),
                    assign_type,
                    compile(state, assign_list),
                ),
            )
        case (Node.LOCAL, name, value):
            return Exec(executors.exec_local_assign, (name, compile(state, value)))
        case (Node.CALL, rule_name, lol):
            return Exec(executors.exec_rule, (compile(state, rule_name), compile(state, lol)))
        case (Node.ACTIONS, *r):
            return Exec(compile_actions, r)
        case (Node.WHILE, cond, block):
            return Exec(executors.exec_while, (cond, compile(state, block)))
        case (Node.FOR, varname, items, block):
            return Exec(
                executors.exec_for,
                (varname, compile(state, items), compile(state, block)),
            )
        case (Node.LOL, *r):
            return (Node.LOL, *[compile(state, item) for item in r])
        case (Node.BREAK, not_used):
            return Exec(executors.exec_break, (not_used,))
        case (Node.CONTINUE, not_used):
            return Exec(executors.exec_continue, (not_used,))
        case (Node.RETURN, val):
            return Exec(executors.exec_return, (compile(state, val),))
        case (Node.IF, *r):
            return compile_if(state, *r)
        case (Node.SWITCH, *r):
            return compile_switch(state, *r)
        case (Node.INCLUDE, arg):
            return Exec(executors.exec_include, (arg,))
        case (Node.ON_TARGET, target, block):
            return Exec(executors.exec_on_target, (target, compile(state, block)))
        case [*items]:
            # block
            res = []
            for item in items:
                exec_item = compile(state, item)
                if exec_item is not None:
                    res.append(exec_item)

            return res
        case _:
            return node


def compile_rule(state: State, name: str, params: tuple | None, block: tuple):
    state.rules[name] = Rule(name, compile(state, params), compile(state, block))


def compile_actions(state: State, flags, name, bindlist, script):
    if name == "Clean":
        return

    if name not in state.rules:
        compile_rule(state, name, [], [])

    action = Actions(name, flags, bindlist, script)
    state.actions[name] = action


def compile_if(state: State, expr, true_block, false_block=None):
    return Exec(
        executors.exec_if,
        (
            compile(state, expr),
            compile(state, true_block),
            compile(state, false_block) if false_block else None,
        ),
    )


def compile_switch(state: State, arg, cases):
    return Exec(
        executors.exec_switch,
        (arg, [(item[1], compile(state, item[2])) for item in cases]),
    )



from jamp.paths import Pathname
from jamp.classes import Vars, State, Exec
from jamp.jam_syntax import Arg, Node
from dataclasses import dataclass
from typing import Union

import itertools
import re

MAGIC_COLON = "\x01"
MAGIC_LEFT = "\x02"
MAGIC_RIGHT = "\x03"


def lol_get(lol: list, idx: int):
    if lol and idx < len(lol):
        return lol[idx]

    return []


def flatten(res: list):
    # try to remove one level
    if isinstance(res, list) and len(res) == 1:
        res = res[0]

    return res


def iter_var(var, skip_empty=True):
    if isinstance(var, str):
        yield var
    elif isinstance(var, list):
        for item in var:
            yield item
    else:
        raise Exception(f"unexpected var for iteration: {var}")


def validate(var):
    if not isinstance(var, list):
        raise Exception(f"validation: expected list, got: {var}")
    if len(var) > 0:
        if not isinstance(var[0], str):
            raise Exception(f"validation: expected str on second level, got: {var[0]}")


def validate_lol(var):
    if not isinstance(var, list):
        raise Exception(f"LOL validation: expected list, got: {var}")
    if len(var) > 0:
        if not isinstance(var[0], list):
            raise Exception(
                f"LOL validation: expected list on second level, got: {var[0]}"
            )


def var_expand(
    var: str, lol: list | None, state_vars: Union[Vars, dict], keep_max=True
):
    """
    var_expand() - variable-expand input string into list of strings

    Would just copy input to output, performing variable expansion,
    except that since variables can contain multiple values the result
    of variable expansion may contain multiple values (a list).  Properly
    performs "product" operations that occur in "$(var1)xxx$(var2)" or
    even "$($(var2))".

    Returns a newly created list.
    """

    # This gets alot of cases: $(<) and $(>)
    if len(var) == 4 and var[0] == "$" and var[1] == "(" and var[3] == ")":
        if (var[2] == "1") or (var[2] == "<"):
            return lol_get(lol, 0)
        if (var[2] == "2") or (var[2] == ">"):
            return lol_get(lol, 1)

    i = var.find("$(")
    if i == -1:
        return [var] if var else []

    prefix = var[:i]
    depth = 1
    i += 2  # skip $(
    inside = ""

    while i < len(var) and depth:
        match var[i]:
            case "(":
                depth += 1
                inside += var[i]
            case ")":
                depth -= 1
                if depth != 0:
                    inside += var[i]
            case ":":
                inside += MAGIC_COLON
            case "[":
                inside += MAGIC_LEFT
            case "]":
                inside += MAGIC_RIGHT
            case _:
                inside += var[i]
        i += 1

    # Recursively expand variable name & rest of input
    variables = var_expand(inside, lol, state_vars) if len(inside) else []

    has_remainder = i < len(var)
    remainder = var_expand(var[i:], lol, state_vars) if has_remainder else []
    vals_out = []

    # Now produce the result chain
    for var in variables:
        varname = var

        # Look for a : modifier in the variable name
        colon_idx = var.find(MAGIC_COLON)
        edits: Edits = None
        if colon_idx > 0:
            varname = var[:colon_idx]
            edits = var_edit_parse(var[colon_idx + 1 :])

        # Look for [x-y] subscripting
        # sub1 is x (0 default)
        # sub2 is length (-1 means forever)
        left_idx = var.find(MAGIC_LEFT)
        sub1 = 0
        sub2 = 0

        subscript = False

        if left_idx > 0:
            subscript = True
            right_idx = varname[left_idx + 1 :].find(MAGIC_RIGHT)
            parts = varname[left_idx + 1 :][:right_idx].split("-")
            sub1 = int(parts[0]) - 1
            if len(parts) == 1:
                sub2 = sub1 + 1
            elif len(parts) == 2 and parts[1]:
                sub2 = int(parts[1])
            else:
                sub2 = -1

            varname = varname[0:left_idx]

        # Get variable value, specially handling $(<), $(>), $(n)
        value = None

        if varname == "<":
            value = lol_get(lol, 0)
        elif varname == ">":
            value = lol_get(lol, 1)
        elif len(varname) == 1 and varname[0] >= "1" and varname[0] <= "9":
            value = lol_get(lol, int(varname) - 1)
        else:
            value = state_vars.get(varname)
            if value is None:
                value = []

        if not isinstance(value, list):
            value = [value]

        if subscript:
            if sub2 == -1:
                value = value[sub1:]
            else:
                value = value[sub1:sub2]

        if not value and edits and edits.empty is not None:
            value = [edits.empty]

        if value and edits and edits.join is not None:
            value = [edits.join.join(value)]

        for val in value:
            if edits:
                if edits.filemods:
                    val = var_edit_file(val, edits)

                    if val == "":
                        continue

                if val and edits.upshift or edits.downshift:
                    val = var_edit_shift(val, edits)

                if edits.quote:
                    val = var_edit_quote(val)

            vals_out.append(val)

    out = []
    product_args = []

    if prefix:
        product_args.append([prefix])

    product_args.append(vals_out)

    if has_remainder:
        product_args.append(remainder)

    if product_args:
        for pr in itertools.product(*product_args):
            val = "".join(pr)
            out.append(val)

    return out


@dataclass
class Edits:
    f: None | Pathname
    parent: bool = False  # :P -- go to parent directory
    filemods: bool = False  # one of the above applied
    downshift: bool = False  # :L -- downshift result
    upshift: bool = False  # :U -- upshift result
    quote: bool = False  # :Q -- quote
    empty: None | str = None  # E -- default for empties
    join: None | str = None  # J -- join list with char
    zeroed: bool = False


# var_edit_parse() - parse : modifiers into PATHNAME structure
#
# The : modifiers in a $(varname:modifier) currently support replacing
# or omitting elements of a filename, and so they are parsed into a
# PATHNAME structure (which contains pointers into the original string).
#
# Modifiers of the form "X=value" replace the component X with
# the given value.  Modifiers without the "=value" cause everything
# but the component X to be omitted.  X is one of:
#
#  G <grist>
#  D directory name
#  B base name
#  S .suffix
#  M (member)
#  R root directory - prepended to whole path
#
# This routine sets:
#
#  f.xxx = None
#   leave the original component xxx
#
#  f.xxx = string
#   replace component xxx with string
#
#  f.xxx = ""
#   omit component xxx
#
# var_edit_file() below and path_build() obligingly follow this convention.
def var_edit_parse(colon_part: str):
    edits = Edits(f=Pathname())

    def strval(part):
        if len(part) == 0 or part[0] != "=":
            return "", 1

        colon_idx = part.find(MAGIC_COLON)
        if colon_idx > 0:
            return part[1:colon_idx], colon_idx + 2

        return part[1:], len(part[1:]) + 2  # skip modificator + '=' + text

    def fileval(colon_part):
        edits.filemods = True

        if not colon_part or colon_part[0] != "=":
            if not edits.zeroed:
                edits.f.zero()
                edits.zeroed = True

            return None, 1

        val, skip = strval(colon_part)
        return val, skip

    i = 0
    while i < len(colon_part):
        skip = 1
        match colon_part[i]:
            case "L":
                edits.downshift = True
            case "U":
                edits.upshift = True
            case "Q":
                edits.quote = True
            case "P":
                edits.parent = True
                edits.filemods = True
            case "@":
                import pdb

                pdb.set_trace()
            case "E":
                edits.empty, skip = strval(colon_part[i + 1 :])
            case "J":
                edits.join, skip = strval(colon_part[i + 1 :])
            case "G":
                edits.f.grist, skip = fileval(colon_part[i + 1 :])
            case "R":
                edits.f.root, skip = fileval(colon_part[i + 1 :])
            case "D":
                edits.f.directory, skip = fileval(colon_part[i + 1 :])
            case "B":
                edits.f.base, skip = fileval(colon_part[i + 1 :])
            case "S":
                edits.f.suffix, skip = fileval(colon_part[i + 1 :])
            case "M":
                edits.f.member, skip = fileval(colon_part[i + 1 :])
            case _:
                # just ignore for now
                # in original it stops here, but I decided to just skip
                # unexpected chars
                pass

        i += skip

    return edits


# var_edit_file() - copy input target name to output, modifying filename
def var_edit_file(path: str, edits: Edits):
    res_path = Pathname()
    res_path.parse(path)

    # Replace any res_path with edits.f
    if edits.f.grist is not None:
        res_path.grist = edits.f.grist

    if edits.f.root is not None:
        res_path.root = edits.f.root

    if edits.f.directory is not None:
        res_path.directory = edits.f.directory

    if edits.f.base is not None:
        res_path.base = edits.f.base

    if edits.f.suffix is not None:
        res_path.suffix = edits.f.suffix

    if edits.f.member is not None:
        res_path.member = edits.f.member

    # if requested, modify res_path to point to parent
    if edits.parent:
        res_path.keep_only_parent()

    # put filename back together
    return res_path.build()


def var_edit_shift(string: str, edits: Edits):
    # Handle upshifting, downshifting now
    if edits.upshift:
        return string.upper()
    elif edits.downshift:
        return string.lower()


def var_edit_quote(string):
    # Handle quoting now
    return string.replace("\\", "\\\\")


def var_string(var: str, lol: list, state_vars: Union[Vars, dict], alone=False):
    res = ""

    i = 0
    while i < len(var):
        while i < len(var) and var[i].isspace():
            res += var[i]
            i += 1

        begin = i
        dollar_found = False
        while i < len(var) and not var[i].isspace():
            if var[i] == "$" and i < len(var) - 1 and var[i + 1] == "(":
                dollar_found = True
            i += 1

        text = var[begin:i]
        if not dollar_found:
            res += text
        elif alone and text == "$(<)":
            res += "<NINJA_SIGIL>out"
        elif alone and text == "$(>)":
            res += "<NINJA_SIGIL>in <NINJA_SIGIL>n"
        else:
            res += " ".join(var_expand(text, lol, state_vars, keep_max=False))

    return res


def expand(state: State, arg: Union[Arg, tuple, str], skip_empty=True):
    """Make list of strings from some type of an argument"""

    if arg is None or (skip_empty and arg == ""):
        return []
    elif arg == "":
        return [""]
    elif isinstance(arg, str):
        res = var_expand(arg, state.params, state.vars)
    elif isinstance(arg, Exec):
        from jamp.executors import Result

        execval = arg.execute(state)
        if execval is None:
            res = []
        elif isinstance(execval, Result):
            res = execval.val
        else:
            raise Exception(f"expected result, got {execval}")
    elif isinstance(arg, Arg):
        res = expand(state, arg.value, skip_empty=skip_empty)
    elif isinstance(arg, list):
        res = []
        for item in arg:
            val = expand(state, item, skip_empty=skip_empty)
            for v in iter_var(val, skip_empty=skip_empty):
                res.append(v)
    else:
        print(type(arg))
        raise Exception(f"could not expand arg: {arg}")

    validate(res)
    return res


def expand_lol(state: State, arg: tuple):
    """Make string from some type of an argument"""

    if isinstance(arg, tuple) and len(arg) and arg[0] == Node.LOL:
        res = []
        for lol_list in arg[1:]:
            res.append(expand(state, lol_list))
    elif isinstance(arg, list):
        res = [[expand(state, item) for item in arg]]
    else:
        raise Exception(f"could not expand LOL: {arg}")

    validate_lol(res)
    return res


def re_expand(state: State, string: str, params: list):
    def exp(m: re.Match[str]) -> str:
        expanded = var_expand(m.string[m.start() : m.end()], params, state.vars)
        return " ".join(expanded)

    return re.sub(r"\$\([^\(]+\)", exp, string)



#
# glob.py - match a string against a simple pattern
# translated from jam/glob.c
#
# Understands the following patterns:
#
# * any number of characters
# ? any single character
# [a-z] any single character in the range a-z
# [^a-z]    any single character not in the range a-z
# \x    match x

BITLISTSIZE = 16


def check_bit(tab: bytearray, bit: int):
    return tab[int(bit / 8)] & (1 << ((bit) % 8))


def match(pat: str, string: str, ignore_case: bool = False):
    i = 0  # pat index
    j = 0  # string index
    while i <= len(pat):
        if i == len(pat):
            # -1 if we have leftovers for string
            return 0 if j >= len(string) else -1

        match pat[i]:
            case "?":
                # any char
                if j >= len(string):
                    return 1
                i += 1
                j += 1
            case "[":
                if j == len(string):
                    return 1

                # scan for matching
                second_idx = pat[i:].find("]")
                if second_idx == -1:
                    return 1

                # build character class bitlist
                bitlist = globchars(pat[i + 1 : second_idx])
                i = second_idx + 1

                if not check_bit(bitlist, ord(string[j])):
                    return 1
                j += 1
            case "*":
                if j == len(string) and (i + 1) == len(pat):
                    return 0

                jj = len(string)  # backwards index for string
                i += 1
                # Try to match the rest of the pattern in a recursive
                # call.  If the match fails we'll back up chars, retrying.
                while jj != j:
                    if i != len(pat):
                        r = match(pat[i:], string[jj:])
                    else:
                        r = 0 if jj == len(string) else -1

                    if r == 0:
                        return 0
                    elif r < 0:
                        return 1

                    jj -= 1
            case "\\":
                i += 1

                # Force literal match of next char.
                if i == len(pat) or string[j] != pat[i]:
                    return 1

                j += 1
            case _:
                if ignore_case and string[j].lower() != pat[i].lower():
                    return 1
                elif j == len(string) or string[j] != pat[i]:
                    return 1

                j += 1
                i += 1


# globchars() - build a bitlist to check for character group match
def globchars(pat: str):
    neg = False
    bitlist = bytearray(BITLISTSIZE)

    i = 0
    if pat[i] == "^":
        neg = True
        i += 1

    while i < len(pat):
        if i + 2 < len(pat) and pat[1] == "-":
            for c in range(pat[0:3]):
                bitlist[ord(c) / 8] |= 1 << (ord(c) % 8)
            i += 3
        else:
            c = ord(pat[i])
            i += 1
            bitlist[int(c / 8)] |= 1 << (c % 8)

    if neg:
        for i in range(len(bitlist)):
            bitlist[i] ^= 0b11111111

    # Don't include \0 in either $[chars] or $[^chars]
    bitlist[0] &= 0b11111110
    return bitlist



#
# Pathname - a name of a file, broken into <grist>dir/base/suffix(member)
#
# <grist> is salt to distinguish between targets that otherwise would
# have the same name:  it never appears in the bound name of a target.
# (member) is an archive member name: the syntax is arbitrary, but must
# agree in parse(), build() and the Jambase.
#
# On VMS, we keep track of whether the original path was a directory
# (without a file), so that $(VAR:D) can climb to the parent.
#

import platform
from pathlib import PurePath
from enum import Enum
from functools import cache


@cache
def check_vms():
    return platform.system() == "OpenVMS"


@cache
def check_windows():
    return platform.system() == "Windows"

@cache
def check_linux():
    return platform.system() == "Linux"


def escape_path(s):
    if check_vms():
        return s.replace("$", "$$").lower()
    return s


def add_paths(s, deps):
    if check_vms():
        for dep in deps:
            s.add(dep.lower())
    else:
        s.update(deps)


class Dir(Enum):
    EMPTY = 0  # empty string */
    DEV = 1  # dev: */
    DEVDIR = 2  # dev:[dir] */
    DOTDIR = 3  # [.dir] */
    DASHDIR = 4  # [-] or [-.dir] */
    ABSDIR = 5  # [dir] */
    ROOT = 6  # [000000] or dev:[000000] */


G_DIR = 0  # take just dir */
G_ROOT = 1  # take just root */
G_VAD = 2  # root's dev: + [abs] */
G_DRD = 3  # root's dev:[dir] + [.rel] */
G_VRD = 4  # root's dev: + [.rel] made [abs] */
G_DDD = 5  # root's dev:[dir] + . + [dir] */

grid = [
    # root/dir	EMPTY	DEV	DEVDIR	DOTDIR	DASH,	ABSDIR	ROOT
    [G_DIR, G_DIR, G_DIR, G_DIR, G_DIR, G_DIR, G_DIR],  # empty
    [G_ROOT, G_DIR, G_DIR, G_VRD, G_VAD, G_VAD, G_VAD],  # dev
    [G_ROOT, G_DIR, G_DIR, G_DRD, G_VAD, G_VAD, G_VAD],  # dev:[dir]
    [G_ROOT, G_DIR, G_DIR, G_DRD, G_DIR, G_DIR, G_DIR],  # dot
    [G_ROOT, G_DIR, G_DIR, G_DRD, G_DDD, G_DIR, G_DIR],  # dash
    [G_ROOT, G_DIR, G_DIR, G_DRD, G_DIR, G_DIR, G_DIR],  # abs
    [G_ROOT, G_DIR, G_DIR, G_VRD, G_DIR, G_DIR, G_DIR],  # root
]


class DirFlags:
    def __init__(self):
        self.flags = Dir.EMPTY
        self.dev = ""
        self.dir = ""

    def __repr__(self):
        return f"{self.flags} {self.dev} {self.dir}"


class Pathname:
    def __init__(self, is_vms=None):
        self.grist = None
        self.root = None
        self.directory = None
        self.base = None
        self.member = None
        self.suffix = None
        self.is_dir = False  # only for VMS, keep track if it's a directory
        self.parent = False

        if is_vms is not None:
            # for testing purposes
            self.is_vms = is_vms
        else:
            self.is_vms = check_vms()

    def zero(self):
        self.grist = ""
        self.root = ""
        self.directory = ""
        self.base = ""
        self.member = ""
        self.suffix = ""
        self.parent = False

    def __repr__(self):
        return (
            f"Path[{self.grist},R={self.root},D={self.directory},B={self.base},"
            f"S={self.suffix},M={self.member}]"
        )

    def parse(self, string):
        idx = string.find(">")

        # check for grist
        if string[0] == "<" and idx > 0:
            self.grist = string[1:idx]
            string = string[idx + 1 :]

        # check for member
        if string.endswith(")"):
            idx = string.rfind("(")
            if idx > 0:
                self.member = string[idx + 1 : -1]
                string = string[0:idx]

        if self.is_vms:
            self.parse_vms(string)
        else:
            path = PurePath(string)
            self.suffix = path.suffix
            self.base = path.stem
            self.root = None
            self.directory = str(path.parent)

    def out_member(self):
        if self.member:
            return "(" + self.member + ")"
        return ""

    def out_grist(self):
        if self.grist:
            if self.grist.startswith("<"):
                return self.grist

            return "<" + self.grist + ">"

        return ""

    def build(self, binding=False):
        if self.is_vms:
            return self.build_vms(binding=binding)

        is_abs = PurePath(self.directory).is_absolute()
        fn = ""
        if self.base:
            fn = self.base + self.suffix
        else:
            fn = self.suffix

        if fn is None:
            fn = ""

        if self.root or self.directory or fn:
            if self.root and self.root != "." and not is_abs:
                path = PurePath(self.root, self.directory, fn)
            else:
                path = PurePath(self.directory, fn)
        else:
            path = ""

        res_path = self.out_grist() + str(path) + self.out_member()
        return res_path

    def keep_only_parent(self):
        if not self.is_vms or (self.is_vms and self.base):
            self.base = None
            self.suffix = None
            self.member = None
        elif self.is_vms:
            self.parent = True

    def parse_vms(self, string):
        # Look for dev:[dir] or dev
        idx = string.find("]")

        if idx > 0:
            self.directory = string[: idx + 1]
        else:
            idx = string.find(":")
            if idx > 0:
                self.directory = string[: idx + 1]

        if idx > 0:
            string = string[idx + 1 :]

        # find suffix
        idx = string.rfind(".")
        if idx > 0:
            self.suffix = string[idx:]
            string = string[0:idx]

        # Leaves base
        self.root = None
        self.base = string

    def dir_flags(self, string: str) -> DirFlags:
        info = DirFlags()

        if string:
            idx = string.find(":")
            if idx > 0:
                info.dev = string[: idx + 1]
                info.dir = string[idx + 1 :]

                if len(info.dir) and info.dir[0] == "[":
                    info.flags = Dir.DEVDIR
                else:
                    info.flags = Dir.DEV
            else:
                info.dev = ""
                info.dir = string

                if string[:2] == "[]":
                    info.flags = Dir.EMPTY
                elif string[:2] == "[.":
                    info.flags = Dir.DOTDIR
                elif string[:2] == "[-":
                    info.flags = Dir.DASHDIR
                else:
                    info.flags = Dir.ABSDIR

        # But if its rooted in any way */
        if info.dir == "[000000]":
            info.flags = Dir.ROOT

        return info

    def build_vms(self, binding=False):
        info_root = self.dir_flags(self.root)
        info_dir = self.dir_flags(self.directory)

        res_path = ""
        g = grid[info_root.flags.value][info_dir.flags.value]

        directory = self.directory or ""
        root = self.root or ""

        if g == G_DIR:
            # take dir
            res_path += directory

        elif g == G_ROOT:
            # take root
            res_path += root

        elif g == G_VAD:
            # root's dev + abs directory
            res_path += info_root.dev + info_dir.dir

        elif g == G_DRD or g == G_DDD:
            # root's dev:[dir] + rel directory
            # sanity checks: root ends with ] */
            if self.root.endswith("]"):
                res_path += root[:-1]
            else:
                res_path += root

            # Add . if separating two -'s
            if g == G_DDD:
                res_path += "."

            # skip [ of dir
            res_path += info_dir.dir[1:]

        elif g == G_VRD:
            # root's dev + rel directory made abs
            res_path += info_root.dev
            res_path += "["

            # skip [. of rel dir */
            res_path += info_dir.dir[2:]

        # Now do the special :P modifier when no file was present.
        # (none)		(none)
        # [dir1.dir2]	[dir1]
        # [dir]		[000000]
        # [.dir]		[]
        # []		[]

        if res_path.endswith("]") and self.parent:
            i = len(res_path) - 1
            prev_c = None
            while i >= 0:
                idx = None
                c = res_path[i]
                if c == ".":
                    idx = i
                elif c == "-":
                    idx = i
                    if i != 0 and res_path[i - 1] == ".":
                        idx = i - 1

                if idx is not None:
                    res_path = res_path[:i] + "]"
                    break

                if c == "[" and prev_c != "]":
                    res_path = res_path[:i] + "[000000]"
                    break

                prev_c = c
                i -= 1

        without_dir = False
        if self.base:
            if not res_path:
                without_dir = True

            res_path += self.base

        # If there is no suffix, we append a "." onto all generated
        # names.  This keeps VMS from appending its own (wrong) idea
        # of what the suffix should be.
        if self.suffix:
            res_path += self.suffix
        elif binding and self.base:
            res_path += "."

        if without_dir and binding:
            res_path = "[]" + res_path

        return self.out_grist() + res_path + self.out_member()



from jamp.pattern import match


def test_patterns():
    assert match("one*", "one") == 0
    assert match("one**", "one") == 0
    assert match("one**o", "oneo") == 0
    assert match("one*a*o", "oneo") == 1
    assert match("?a", "da") == 0
    assert match("d?n", "d$n") == 0
    assert match("?aa", "da") == 1
    assert match("?a", "daa") == -1
    assert match("a??", "da") == 1
    assert match("[abd]c", "ac") == 0
    assert match("[abe]c", "dc") == 1
    assert match("[^abd]c", "dc") == 1
    assert match("[^abd]c", "ac") == 1
    assert match("ac*", "aca") == 0
    assert match("ac*d", "aca123d") == 0
    assert match("ac*c", "aca123d") == 1
    assert match("*c*c", "1ca123d") == 1
    assert match("*c*c", "1ca123c") == 0
    assert match("*c*[efg]?", "1ca123cfi") == 0
    assert match("*c*[efg]j", "1ca123cfi") == 1
    assert match("*c*[^efg]j", "1ca123ckj") == 0



from jamp.jam_syntax import parse


def test_simple_assign_sequence():
    rule_t = """
    rule One
    {
        a = 1 ;
        b = 2 ;
        c = 3 ;
    }

    rule Two
    {
        c = 1 ;
        d = 2 ;
        e = 3 ;
    }
    """
    ast = parse(rule_t)
    assert str(ast) == (
        "[(RULE, 'One', None, [(ASSIGN, a, '=', [1]), (ASSIGN, b, '=', [2]), "
        "(ASSIGN, c, '=', [3])]), (RULE, 'Two', None, [(ASSIGN, c, '=', [1]),"
        " (ASSIGN, d, '=', [2]), (ASSIGN, e, '=', [3])])]"
    )


def test_rule_if_with_call():
    rule_t = """
    rule R1
    {
        C1 $(1) ;

        if $(a) = .$(ext)
        {
            NOUPDATE $(a) ;
        }
        C2 $(<) : $(>) : : $(<:S=.$(L)) ;
        local V = [ C3 $(>:S=.$(O)) ] ;
    }
    """
    ast = parse(rule_t)
    assert str(ast) == (
        "[(RULE, 'R1', None, "
        "[(CALL, C1, (LOL, [$(1)])), "
        "(IF, (EXPR_BOP, ((EXPR, $(a)), '=', (EXPR, .$(ext)))), "
        "[(CALL, NOUPDATE, (LOL, [$(a)]))]), "
        "(CALL, C2, (LOL, [$(<)], [$(>)], None, [$(<:S=.$(L))])), "
        "(LOCAL, [V], [(FUNC, C3, (LOL, [$(>:S=.$(O))]))])])]"
    )


def test_rule_with_local_and_assign():
    rule_t = """
    rule One
    {
        local a = 1 ;
        b = 2 ;
        local c  ;
    }
    """

    ast = parse(rule_t)
    assert (
        str(ast)
        == "[(RULE, 'One', None, [(LOCAL, [a], [1]), (ASSIGN, b, '=', [2]), (LOCAL, [c], None)])]"
    )


def test_rule_with_params():
    rule_t = """
    rule One
    {
        local a = 1 ;
        b = 2 ;
        local c  ;
    }
    """

    ast = parse(rule_t)
    assert (
        str(ast)
        == "[(RULE, 'One', None, [(LOCAL, [a], [1]), (ASSIGN, b, '=', [2]), (LOCAL, [c], None)])]"
    )


def test_args():
    patterns = [
        ("a = $(a) ;", "[(ASSIGN, a, '=', [$(a)])]"),
        ('a = $(a;J=" ") ;', "[(ASSIGN, a, '=', [$(a;J= )])]"),
        ('a ?= "some text" ;', "[(ASSIGN, a, '?=', [some text])]"),
        ('a += "some text" ;', "[(ASSIGN, a, '+=', [some text])]"),
        (
            "b default = $(a) $(b) ;",
            "[(ASSIGN, b, 'default', [$(a), $(b)])]",
        ),
    ]

    for text, expected in patterns:
        parsed = parse(text)
        assert str(parsed) == expected


def test_call():
    patterns = [
        ("LINK some ;", "[(CALL, LINK, (LOL, [some]))]"),
    ]

    for text, expected in patterns:
        parsed = parse(text)
        assert str(parsed) == expected


def test_block():
    text = """
    {
        Echo "1" ;
    }
    """

    ast = parse(text)
    assert str(ast) == "[[(CALL, Echo, (LOL, [1]))]]"


def test_if_with_block():
    text = """
    if $(a)
    {
        Echo "1" ;
    }
    """

    ast = parse(text)
    assert str(ast) == "[(IF, (EXPR, $(a)), [(CALL, Echo, (LOL, [1]))])]"


def test_if_with_else():
    """the third Echo should be outside of if block"""
    text = """
    if $(a)
    {
        Echo "1" ;
    }
    else
        Echo "2" ;
        Echo "3" ;
    """

    ast = parse(text)
    assert (
        str(ast)
        == "[(IF, (EXPR, $(a)), [(CALL, Echo, (LOL, [1]))], (CALL, Echo, (LOL, [2]))),"
        " (CALL, Echo, (LOL, [3]))]"
    )


def test_actions():
    patterns = [
        (
            'actions one  { echo "1" }',
            """[(ACTIONS, [], 'one', (BINDLIST, None), ' echo "1" ')]""",
        ),
    ]

    for text, expected in patterns:
        parsed = parse(text)
        assert str(parsed) == expected


def test_switch():
    text = """
    a = "test1" ;
    switch $(a)
    {
        case ?est1 :
            Echo "test1" ;
        case ?est2 :
            Echo "test2" ;
        case ?est3 :
            Echo "test3" ;
    }
    """

    ast = parse(text)
    assert str(ast) == (
        "[(ASSIGN, a, '=', [test1]), "
        "(SWITCH, [$(a)], ["
        "(CASE, '?est1', [(CALL, Echo, (LOL, [test1]))]), "
        "(CASE, '?est2', [(CALL, Echo, (LOL, [test2]))]), "
        "(CASE, '?est3', [(CALL, Echo, (LOL, [test3]))])])]"
    )


def test_arg_on_target():
    text = """
    rule Object
    {
        Clean clean : $(<) ;
        SEARCH on $(>) = $(SEARCH_SOURCE) ;
    }
    """
    ast = parse(text)
    assert str(ast) == (
        "[(RULE, 'Object', None, "
        "[(CALL, Clean, (LOL, [clean], [$(<)])), "
        "(ARG_ON_TARGET, SEARCH, [$(>)], '=', [$(SEARCH_SOURCE)])])]"
    )


def test_on_target():
    text = """
    a on test.c = "val1" ;
    b on test2.c = "val3" ;
    c = "val2" ;

    on test.c Echo $(a) $(b) $(c) ;
    on test2.c Echo $(a) $(b) $(c) ;
    """
    ast = parse(text)
    assert (
        "[(ARG_ON_TARGET, a, [test.c], '=', [val1]), "
        "(ARG_ON_TARGET, b, [test2.c], '=', [val3]), "
        "(ASSIGN, c, '=', [val2]), "
        "(ON_TARGET, test.c, (CALL, Echo, (LOL, [$(a), $(b), $(c)]))), "
        "(ON_TARGET, test2.c, (CALL, Echo, (LOL, [$(a), $(b), $(c)])))]"
    ) == str(ast)


def test_func():
    text = """
    rule one
    {
        a = 1 2 ;
        return $(1) $(a) ;
    }

    res = [ one "one" ] ;
    Echo $(res) ;
    """
    ast = parse(text)
    assert str(ast) == (
        "[(RULE, 'one', None, "
        "[(ASSIGN, a, '=', [1, 2]), "
        "(RETURN, [$(1), $(a)])]), "
        "(ASSIGN, res, '=', [(FUNC, one, (LOL, [one]))]), "
        "(CALL, Echo, (LOL, [$(res)]))]"
    )


def test_parse_assign_empty():
    text = """
    a = "" ;
    b = ;
    """
    ast = parse(text)
    assert str(ast) == """[(ASSIGN, a, '=', [""]), (ASSIGN, b, '=', None)]"""


def test_parse_assign_optional():
    text = """
    C++FLAGS ?= /NAMES=UPPER ;
    """
    ast = parse(text)
    assert str(ast) == "[(ASSIGN, C++FLAGS, '?=', [/NAMES=UPPER])]"


def test_on_arg_return():
    text = """
    Archive test.a : test.obj ;
    AR on test.a = "ar ru" ;
    Echo [ on test.a return $(AR) ] ;
    """
    ast = parse(text)
    assert str(ast) == (
        "[(CALL, Archive, (LOL, [test.a], [test.obj])), "
        "(ARG_ON_TARGET, AR, [test.a], '=', [ar ru]), "
        "(CALL, Echo, (LOL, [(RETURN_ON, test.a, [$(AR)])]))]"
    )

def test_parse_lol_with_empty():
    text = """
    One 1 : : 2 ;
    """
    ast = parse(text)
    assert str(ast) == "[(CALL, One, (LOL, [1], None, [2]))]"



#!/usr/bin/env python3

''' create test.h with some delay '''

import sys
file_c = sys.argv[1]
file_h = sys.argv[2]
file_in = sys.argv[3]

with open(file_in, 'r') as fin:
    header = ''
    source = ''
    res = []
    for line in fin.readlines():
        if line.startswith('--'):
            if res:
                header = '\n'.join(res)

            res = []
        else:
            res.append(line)

    source = ''.join(res)

    with open(file_h, 'w') as fheader:
        with open(file_c, 'w') as fsource:
            fheader.write(header)
            fsource.write(source)
